/*
 * Copyright (C) 2009, Stefan Horn
 * 
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for
 * more details.
 *
 * Layout is based on Alexander Kudjashev <Kudjashev@gmail.com>,
 * James Simmons, Geert Uytterhoeven and arcfb.c by Jaya Kumar.
 * 
 * Framebuffer for LS020... based LCD Controller with 176x132 Pixel
 * Thanks Christian Kranz for the Great Job
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/kthread.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>
#include <linux/uaccess.h>
#include <mach/gpio.h>


/* lcd resolution */
#define X_RES       176
#define Y_RES       132
#define B_PP        16
#define RedBits	    5
#define GreenBits   6
#define BlueBits    5

#define MEM_LEN     (X_RES * Y_RES * B_PP / 8)
#define BLOCK_LEN   3872
#define REFR_TIME   (HZ / 10)

#define CmdPin      AT91_PIN_PB8
#define RstPin      AT91_PIN_PB9

struct ls020_par {
	struct task_struct  *ls020_thread_task;
	struct spi_device   *spi;
	u8                  *buffer;
	u8                  *shadow;
};

static void spi_command(struct spi_device *spi, unsigned int c)
{
	unsigned int w = c;
	at91_set_gpio_value(CmdPin, 1);
	ndelay(5);
	spi_write(spi, (u8 *)&w, 2);
    	at91_set_gpio_value(CmdPin, 0);
	ndelay(5);
}

static void spi_data(struct spi_device *spi, unsigned int d)
{
	unsigned int w = d;
	at91_set_gpio_value(CmdPin, 0);
	ndelay(5);
	spi_write(spi, (u8 *)&w, 2);
}

static struct fb_fix_screeninfo ls020_fix __devinitdata = {
	.id             = "ls020",
	.type           = FB_TYPE_PACKED_PIXELS,
	.visual         = FB_VISUAL_TRUECOLOR,
	.xpanstep       = 0,
	.ypanstep       = 0,
	.ywrapstep      = 0,
	.line_length    = X_RES * B_PP / 8,
	.accel          = FB_ACCEL_NONE,
};

static struct fb_var_screeninfo ls020_var __devinitdata = {
	.xres           = X_RES,
	.yres           = Y_RES,
	.xres_virtual   = X_RES,
	.yres_virtual   = Y_RES,
	.height         = -1,
	.width          = -1,
	.vmode          = FB_VMODE_NONINTERLACED,
	.bits_per_pixel = B_PP,
	.red            = { 11, 5, 0 },
	.green          = { 5, 6, 0 },
	.blue           = { 0, 5, 0 },
	.nonstd         = 0,
};

static void spi_complete(void *param)
{

}

static int ls020_blank(int blank_mode, struct fb_info *info)
{
//	struct ls020_par *par = info->par;

	switch (blank_mode) {
	case FB_BLANK_UNBLANK:
//		spi_command(par->spi, DISPON);
//		break;
	case FB_BLANK_NORMAL:
	case FB_BLANK_VSYNC_SUSPEND:
	case FB_BLANK_HSYNC_SUSPEND:
	case FB_BLANK_POWERDOWN:
//		spi_command(par->spi, DISPOFF);
//		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int ls020_mmap(struct fb_info *info, struct vm_area_struct *vma)
{
	unsigned long size, offset;

	size = vma->vm_end - vma->vm_start;
	offset = vma->vm_pgoff << PAGE_SHIFT;

	offset += info->fix.smem_start;
	if (remap_pfn_range(vma, vma->vm_start, offset >> PAGE_SHIFT,
			    size, vma->vm_page_prot))
		return -EAGAIN;

	printk(KERN_INFO "ls020: mmap framebuffer P(%lx)->V(%lx), size %lx\n",
		offset, vma->vm_start, size);
	return 0;
}

static int ls020_fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
			 u_int transp, struct fb_info *info)
{
	if (regno >= 256)
		return -EINVAL;

	if (info->var.grayscale) {
		/* grayscale = 0.30*R + 0.59*G + 0.11*B */
		red = green = blue =
		    (red * 77 + green * 151 + blue * 28) >> 8;
	}

	if (info->fix.visual == FB_VISUAL_TRUECOLOR) {

		u32 value;
		/* Place color in the pseudopalette */
		if (regno > 16)
			return -EINVAL;

		red   >>= (16 - info->var.red.length);
		green >>= (16 - info->var.green.length);
		blue  >>= (16 - info->var.blue.length);

		value = (red   << info->var.red.offset) |
			(green << info->var.green.offset)|
			(blue  << info->var.blue.offset);
		value &= 0xFFFF;

		((u32 *) (info->pseudo_palette))[regno] = value;
	}
	return 0;
}


static struct fb_ops ls020_ops = {
	.owner          = THIS_MODULE,
	.fb_read        = fb_sys_read,
	.fb_write       = fb_sys_write,
	.fb_fillrect    = sys_fillrect,
	.fb_copyarea    = sys_copyarea,
	.fb_imageblit   = sys_imageblit,
	.fb_blank       = ls020_blank,
	.fb_mmap        = ls020_mmap,
	.fb_setcolreg   = ls020_fb_setcolreg,
};

static int ls020_thread(void *param)
{
	struct ls020_par *par = (struct ls020_par *)param;
	int x, i;

	set_user_nice(current, 5); 

	while(!kthread_should_stop()) {
		/* Optimize refresh: transfer image only when it has changed */
		if(memcmp(par->shadow, par->buffer, MEM_LEN)) {
			memcpy(par->shadow, par->buffer, MEM_LEN);	
			/* send frame */
			spi_command(par->spi, 0xef90);
			spi_command(par->spi, 0x0504);
			spi_command(par->spi, 0x0800);
			spi_command(par->spi, 0x0983);
			spi_command(par->spi, 0x0a00);
			spi_command(par->spi, 0x0baf);
			/* rotate */
			spi_command(par->spi, 0xef80);
			spi_command(par->spi, 0x1806);
			spi_command(par->spi, 0x1200);
			spi_command(par->spi, 0x1583);
			spi_command(par->spi, 0x13af);
			spi_command(par->spi, 0x1600);

			for(i = 0; i < MEM_LEN; i += BLOCK_LEN) {
				/* write some stuff */	
				spi_write(par->spi, &par->buffer[i], BLOCK_LEN);
			}
		}
		schedule_timeout_interruptible(REFR_TIME);
	}

	return 0;
}

static int __devinit ls020_probe(struct spi_device *spi)
{
	struct fb_info *info;
	int retval;
	struct ls020_par *par;

	spi->master->bus_num = 1;
	spi->chip_select = 0;
	spi->max_speed_hz = 8 * 1000 * 1000;
	spi->mode = SPI_MODE_3;
	spi->bits_per_word = 16;

	retval = spi_setup(spi);
	if(retval < 0)
		return retval;	

		
	at91_set_GPIO_periph(CmdPin, 1);
	gpio_direction_output(CmdPin,1);
	at91_set_GPIO_periph(RstPin, 1);
	gpio_direction_output(RstPin,1);
	

	/* Reset the Display */
	at91_set_gpio_value(RstPin, 0);
	mdelay(10);
	at91_set_gpio_value(RstPin, 1);
	mdelay(50);

	/* LCD init 1*/
	spi_command(spi, 0xfdfd);
	spi_command(spi, 0xfdfd);
	spi_command(spi, 0xef00);
	spi_command(spi, 0xee04);
	spi_command(spi, 0x1b04);
	spi_command(spi, 0xfefe);
	spi_command(spi, 0xfefe);
	spi_command(spi, 0xef90);
	spi_command(spi, 0x4a04);
	spi_command(spi, 0x7f3f);
	spi_command(spi, 0xee04);
	spi_command(spi, 0x4306);

	mdelay(7);

	/* LCD init 2*/
	spi_command(spi, 0xef90);
	spi_command(spi, 0x0983);
	spi_command(spi, 0x0800);
	spi_command(spi, 0x0baf);
	spi_command(spi, 0x0a00);
	spi_command(spi, 0x0500);
	spi_command(spi, 0x0600);
	spi_command(spi, 0x0700);
	spi_command(spi, 0xef00);
	spi_command(spi, 0xee0c);
	spi_command(spi, 0xef90);
	spi_command(spi, 0x0080);
	spi_command(spi, 0xefb0);
	spi_command(spi, 0x4902);
	spi_command(spi, 0xef00);
	spi_command(spi, 0x7f01);
	spi_command(spi, 0xe181);
	spi_command(spi, 0xe202);
	spi_command(spi, 0xe276);
	spi_command(spi, 0xe183);
	spi_command(spi, 0x8001);
	spi_command(spi, 0xef90);
	spi_command(spi, 0x0000);
	
	retval = -ENOMEM;
		
	info = framebuffer_alloc(sizeof(struct ls020_par), &spi->dev);
	if(!info)
		return retval;

	info->screen_base = alloc_pages_exact(MEM_LEN, GFP_DMA | __GFP_ZERO);
	if(!info->screen_base)
		goto err;

	info->fbops = &ls020_ops;
	info->var = ls020_var;
	info->fix = ls020_fix;

	info->fix.smem_len = MEM_LEN;
	info->fix.smem_start = (unsigned long)virt_to_phys(info->screen_base);
	info->screen_size = info->fix.smem_len;
	info->flags = FBINFO_FLAG_DEFAULT;

	//pseudo palette
	info->pseudo_palette = alloc_pages_exact(MEM_LEN, GFP_DMA | __GFP_ZERO);
	if(!info->pseudo_palette)
		goto err1;
	//cmap
	if (fb_alloc_cmap(&info->cmap, 256, 0) < 0) {
		goto err2;
	}

	par = info->par;
	par->spi = spi;
	par->buffer = info->screen_base;

	par->shadow = kmalloc(MEM_LEN, GFP_KERNEL | GFP_DMA);
	if(!par->shadow)
		goto err2;

	retval = register_framebuffer(info);
	if (retval < 0)
		goto err3;

	dev_set_drvdata(&spi->dev, info);

	par->ls020_thread_task = kthread_run(ls020_thread, par, "ls020");

	retval = -EIO;
	if(IS_ERR(par->ls020_thread_task))
		goto err3;
	
	printk(KERN_INFO "fb%d: %s frame buffer device, %lx base address %dK of video memory\n",
	       info->node, info->fix.id, info->fix.smem_start, info->fix.smem_len >> 10);

	return 0;

err3:
	kfree(par->shadow);
err2:
	free_pages_exact(info->pseudo_palette, MEM_LEN);
err1:
	free_pages_exact(par->buffer, MEM_LEN);
err:
	framebuffer_release(info);

	return retval;
}

static int __devexit ls020_remove(struct spi_device *spi)
{
	struct fb_info *info = dev_get_drvdata(&spi->dev);

	if (info) {
		struct ls020_par *par = info->par;

		unregister_framebuffer(info);
		free_pages_exact(par->buffer, MEM_LEN);
		kfree(par->shadow);
		framebuffer_release(info);

		kthread_stop(par->ls020_thread_task);
	}
	return 0;
}

static struct spi_driver ls020_driver = {
	.driver = {
		.name       = "ls020",
		.bus        = &spi_bus_type,
		.owner      = THIS_MODULE,
	},
	.probe      = ls020_probe,
	.remove     = __devexit_p(ls020_remove),
};

static int __init ls020_init(void)
{
	printk("ls020 spi fb driver\n");
	return spi_register_driver(&ls020_driver);
}

static void __exit ls020_exit(void)
{
	spi_unregister_driver(&ls020_driver);
}

module_init(ls020_init);
module_exit(ls020_exit);

MODULE_AUTHOR("Alexander Kudjashev");
MODULE_DESCRIPTION("ls020 spi fb driver");
MODULE_LICENSE("GPL");
